{Program przetlumaczony na Pascala z
 "Ray tracer in F# (C) Flying Frog Consultancy Ltd., 2007"
 http://www.ffconsultancy.com/dotnet/fsharp/raytracer/index.html }
program raytracer_workers;
uses SysUtils, SDL, math, matrix, UComplex, FGL;

const
   DELTA = 1E-6;
   RATIO = 1/3;
   WIDTH = 800;
   HEIGHT = 600;
   LEVELS = 5;
         
function Quitted : Boolean;
var event : TSDL_Event;
begin
   Quitted := False;
   while SDL_PollEvent (@event) = 1
   do case event.type_ of
      SDL_KeyDown:
         Quitted := Quitted or (event.key.keysym.sym = LongWord ('q'));
      SDL_QuitEv: Quitted := True;
   end;
end;

type
   TColor = Tvector3_double;
   TVec = Tvector3_double;
   TMatrix = Tmatrix4_double;
   
   function TrVec (const R : TVec) : Tvector4_double;
   {Translatable vector.}
   begin
      result := Tvector3_double (R);
      result.data[3] := 1
   end;
   function Translate (const R : TVec) : TMatrix;
   begin
      result.Init_Identity;
      result.Set_Column (3, R);
      result.data[3,3] := 1;
   end;
   function RotX (t : Double) : TMatrix;
   begin
      RotX.Init (1, 0, 0, 0,
                 0, cos (t), sin (t), 0,
                 0, -sin (t), cos (t), 0,
                 0, 0, 0, 1)
   end;
   function RotY (t : Double) : TMatrix;
   begin
      RotY.Init (cos (t), 0, sin (t), 0,
                 0, 1, 0, 0,
                 -sin (t), 0, cos (t), 0,
                 0, 0, 0, 1)
   end;
   function RotZ (t : Double) : TMatrix;
   begin
      RotZ.Init (cos (t), sin (t), 0, 0,
                 -sin (t), cos (t), 0, 0,
                 0, 0, 1, 0,
                 0, 0, 0, 1)
   end;
   
type
   TMaterial = record
      Color : TColor;
      Shininess : Double
   end;
   TSphere = record
      Center : TVec;
      Radius : Double
   end;
   TObjKind = (Sphere, Group);
   PObj = ^TObj;
   TObjList = specialize TFPGList<PObj>;
   TObj = record
      BoundingBall : TSphere;
      case Kind : TObjKind of
         Sphere: (Material : TMaterial);
         Group: (GroupObjs : TObjList);
   end;
   TRay = record
      Origin : TVec;
      Direction : TVec;
   end;
   THit = record
      Distance : Double;
      Normal : TVec;
      ObjMat : TMaterial
   end;
   
   function SkyMaterial : TMaterial;
   begin
      result.Color.Init (0, 0.1, 0.2);
      result.Shininess := 0
   end;
   
   function FloorMaterial (x, y : Double) : TMaterial;
   var z : Complex; c1, c2 : TColor;
   begin
      z.re := x; z.im := y;
      c1.Init (0.7, 1, 0.5); c2.Init (0.2, 0.3, 0.1);
      if Trunc (cmod (z) + carg (z) / pi) mod 2 = 0
      then begin
         result.Color := c1 * (1/(1+(y+2)*(y+2))) +
            c2 * (1 - abs(1/(1+(y+2)*(y+2))));
         result.Shininess := 0.25
      end else begin
         result.Color := c2;
         result.Shininess := 0
      end
   end;
      
   function SphereMaterial (n : Integer) : TMaterial;
   begin
      result.Color.Init ((1 - 0.5*n/LEVELS)*((LEVELS-n) mod 2),
                         (1 - 0.5*n/LEVELS)*((LEVELS-n+1) mod 2), 0);
      result.Shininess := 0.5
   end;
      
   function IntersectSphere (const Ray : TRay; const Sphere : TSphere) : Double;
   var V : TVec; B2, Discr, T1, T2 : Double;
   begin
      V := Ray.Origin - Sphere.Center;
      B2 := Ray.Direction ** V;
      {The direction is normalized.}
      Discr := B2*B2 - (V.Squared_Length - Sphere.Radius*Sphere.Radius);
      if Discr < 0 then exit (Infinity);
      Discr := Sqrt (Discr);
      T2 := -B2 + Discr;
      if T2 <= 0
      then result := Infinity
      else begin
         T1 := -B2 - Discr;
         if T1 > 0 then result := T1
         else result := T2
      end;
   end;
   
   function IntersectSky : THit;
   begin
      result.Distance := Infinity;
      result.Normal.Init (0, -1, 0);
      result.ObjMat := SkyMaterial
   end;
   
   procedure IntersectFloor (const Ray : TRay; var Hit : THit);
   var P : TVec; HitDistance : Double;
   begin
      HitDistance := -Ray.Origin.data[1] / Ray.Direction.data[1];
      if (HitDistance <= 0) or (HitDistance > Hit.Distance)
      then exit;
      Hit.Distance := HitDistance;
      Hit.Normal.Init (0, 1, 0);
      P := Ray.Origin + Ray.Direction * HitDistance;
      Hit.ObjMat := FloorMaterial (P.data[0], P.data[2])
   end;
   
   procedure Intersect (const Ray : TRay; const Scene : TObj; var Hit : THit);
   var
      HitDistance : Double;
      Obj : ^TObj;
   begin
      HitDistance := IntersectSphere (Ray, Scene.BoundingBall);
      if HitDistance >= Hit.Distance then exit;
      case Scene.Kind of
         Sphere:
            begin
               Hit.Normal := Ray.Origin + Ray.Direction * HitDistance
                  - Scene.BoundingBall.Center;
               Hit.Normal /= Hit.Normal.Length;
               Hit.ObjMat := Scene.Material;
               Hit.Distance := HitDistance
            end;
         Group:
            for Obj in Scene.GroupObjs do Intersect (Ray, (Obj^), Hit);
      end
   end;
   
   function RayTrace (const LightDir : TVec; const Scene : TObj;
                      X, Y : Integer) : TColor;
   var
      Ray : TRay;   
   
      function Reflect (Weight : Double) : TColor;
      {We use Ray as loop variable.}
      var
         Hit, ShadowHit : THit; P : TVec;
         Color : TColor; refl, light : Double;
      begin
         Hit := IntersectSky;
         IntersectFloor (Ray, Hit);
         Intersect (Ray, Scene, Hit);
         if Hit.Distance = Infinity then
         begin                  {Skyline}
            result.Init (0, Sqrt (Ray.Direction.data[1]), 1); exit
         end;
         P := Ray.Origin + Ray.Direction * Hit.Distance + Hit.Normal * DELTA;
         {Recursively examine specular reflections.}
         if Weight < 0.1 then Color.Init (0.5, 0.5, 0.5)
         else begin
            Ray.Origin := P;
            Ray.Direction -= Hit.Normal * (Ray.Direction ** Hit.Normal * 2);
            {Ray.Direction /= Ray.Direction.Length; -- preserved}
            refl := Hit.ObjMat.Shininess;
            Color := Reflect (Weight * refl) * refl;
         end;
         
         {Ray not needed, reuse to detect shadow.}
         Ray.Origin := P; Ray.Direction := -LightDir;
         ShadowHit := IntersectSky;
         Intersect (Ray, Scene, ShadowHit);
         if ShadowHit.Distance = Infinity
         then light := max (0, - (Hit.Normal ** LightDir))
         else light := 0.;
         Reflect := Hit.ObjMat.Color * (Color + light)
      end;
   begin
      Ray.Origin.Init (0, 2, -6);
      Ray.Direction.Init (X - 0.5*WIDTH, Y - 0.5*HEIGHT, max (WIDTH, HEIGHT));
      Ray.Direction /= Ray.Direction.Length;
      Ray.Direction := RotX (-0.1) * TrVec (Ray.Direction);
      result := Reflect (1.0)
   end;
      
   procedure BoundingSphere (var Bound : TSphere; const Scene : TObj);
   var SubObj : ^TObj;
   begin
      Bound.Radius :=
         max (Bound.Radius,
              Scene.BoundingBall.Radius +
                 (Bound.Center - Scene.BoundingBall.Center).Length);
      case Scene.Kind of
         Sphere: ;
         Group: for SubObj in Scene.GroupObjs do
                   BoundingSphere (Bound, (SubObj^));
      end
   end;
   
   var TrZERO : Tvector4_double;
   
   function BuildScene
      (Level : Integer; LRadius : Double; const Move : TMatrix) : PObj;
   var
      SphereObj : ^TObj;
      Offset : TVec;
      function MakeTop (i : Integer) : TMatrix;
      begin
         result := Move * RotY (pi / 6 + 2*pi/3 * i)
            * RotZ (pi / 4)
            * Translate (Offset)
      end;
      function MakeBottom (i : Integer) : TMatrix;
      begin
         result := Move * RotY (pi / 3 * i)
            * RotZ (110 * pi / 180)
            * Translate (Offset)
      end;
   var i : Integer;
   begin
      Offset.Init_zero; Offset.data[1] := (1 + RATIO) * LRadius;
      new (SphereObj);
      SphereObj^.Kind := Sphere;
      SphereObj^.BoundingBall.Center := Move * TrZERO;
      SphereObj^.BoundingBall.Radius := LRadius;
      SphereObj^.Material := SphereMaterial (Level);
      if Level <= 1 then exit (SphereObj);
      new (result);
      result^.BoundingBall := SphereObj^.BoundingBall;
      result^.Kind := Group;
      result^.GroupObjs := TObjList.Create;
      result^.GroupObjs.Add(SphereObj);
      for i := 0 to 2 do
         result^.GroupObjs.Add
            (BuildScene (Level-1, LRadius*RATIO, MakeTop (i)));
      for i := 0 to 5 do
         result^.GroupObjs.Add
            (BuildScene (Level-1, LRadius*RATIO, MakeBottom (i)));
      BoundingSphere (result^.BoundingBall, result^ );
   end;
   
   procedure Destroy (var Obj : PObj);
   var SubObj : PObj;
   begin
      if Obj = nil then exit;
      case Obj^.Kind of
         Sphere: ;
         Group: begin
                   for SubObj in Obj^.GroupObjs do Destroy (SubObj);
                   Obj^.GroupObjs.Destroy;
                end;
      end;
      dispose (Obj);
      Obj := nil
   end;
   
var   
   screen : PSDL_Surface;
   
   procedure PutPixel (const x, y : Integer; const Color : TColor);
   var
      PixelColor : LongWord;
      PixelLocation : ^LongWord;
      c : array[0..2] of Byte; i : Integer;
   begin
      for i := 0 to 2 do
         c[i] := EnsureRange (round (0.5 + 255.0 *  Color.data[i]), 0, $ff);
      PixelColor := SDL_MapRGB (screen^.format, c[0], c[1], c[2]);
      PixelLocation := screen^.pixels +
         y * screen^.pitch + x * screen^.format^.BytesPerPixel;
      PixelLocation^ := PixelColor;
   end;
var
   Scene : PObj;
   LightDir : TVec;
type
   TRayJob = class
      Line: array[0..WIDTH-1] of TColor;
      Y: Integer;
      constructor Create (aY: Integer);
      procedure Execute;
      procedure Report;
      function Info: String;
   end;
   constructor TRayJob.Create (aY: Integer);
   begin
      WriteLn ('Creating job ',aY);
      Y := aY;
   end;
   procedure TRayJob.Execute;
   var X: Integer;
   begin
      WriteLn ('Executing job ', Y);
      for X := 0 to WIDTH-1 do Line[X] := RayTrace (LightDir, Scene^, X, Y);
   end;
   procedure TRayJob.Report;
   var X: Integer;
   begin
      WriteLn ('Reporting job ',Y);
      for X := 0 to WIDTH-1 do PutPixel (X, HEIGHT-Y-1, Line[X]);      
      SDL_Flip (screen);
   end;
   function TRayJob.Info: String;
   begin
      Result := 'Job ' + IntToStr (Y);
   end;
type
   TJobQueue = specialize TFPGObjectList<TRayJob>;
var
   Y : Integer;
   InitMoveVec : TVec;
   Manager: TJobQueue;
   Job: TRayJob;
type SDL_Exn = class (Exception);
begin
   TrZERO.Init_zero; TrZERO.data[3] := 1;
   Scene := nil;
   Manager := TJobQueue.Create;
   try try
      if SDL_Init(SDL_INIT_VIDEO) < 0 then
         raise SDL_Exn.Create ('Couldn''t initialize SDL');
      screen := SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_SWSURFACE);
      if screen = nil then
         raise SDL_Exn.Create ('Couldn''t set video mode');
      
      LightDir.Init (-1, -3, 2);
      LightDir /= LightDir.Length;
      InitMoveVec.Init (0, 1, 0);
      Scene := BuildScene (LEVELS, 1, Translate (InitMoveVec));
      for Y := 0 to HEIGHT-1 do Manager.Add (TRayJob.Create (Y));
      for Job in Manager do Job.Execute;
      for Job in Manager do Job.Report;      
      Manager.Destroy;
      {SDL_Flip (screen);}
      {repeat SDL_Delay (20) until Quitted}
   except
      on E: SDL_Exn do
         WriteLn ('ERROR: ', E.Message, ' -- ', SDL_GetError);
   end
   finally
      Destroy (Scene);
      if screen <> nil then SDL_FreeSurface(screen);
      SDL_Quit
   end
end.   
