unit workers_asynch;
{ Job queue manager with fixed number of worker threads. }

interface

uses {$ifdef unix}
  cthreads,
  {cmem,} // the c memory manager is on some systems much faster for
        // multi-threading
  {$endif}
  SysUtils, Classes, FGL;

const
   WORKER_IDLETIME = 100;
type
   TJob = procedure of object;
   TJobQueue = specialize TFPGList<TJob>;
   TWorkManager = class;
   
   TWorker = class (TThread)
   private
      Manager: TWorkManager;
   protected
      procedure Execute; override;
   public
      constructor Create (aManager: TWorkManager);
   end;
   TWorkerList = specialize TFPGList<TWorker>;
   
   TWorkManager = class
   private
      ManagerLock: TRTLCriticalSection;
      JobQueue: TJobQueue;
      WorkerList: TWorkerList;
      GotJob: PRTLEvent;
      FJobs: Integer;
      function TakeJob : TJob;
   public
      ExecutedJob: PRTLEvent;
      constructor Create (NumWorkers: Integer);
      procedure AddJob (Job: TJob);
      destructor Destroy; override;
      property Jobs: Integer read FJobs;
      { procedure Suspend;
        procedure Resume; etc...}
   end;
   
implementation

constructor TWorker.Create (aManager: TWorkManager);
begin
   Manager := aManager;
   FreeOnTerminate := True;
   inherited Create (False); { start rightaway }
end;

procedure TWorker.Execute;
var CurrentJob : TJob;
begin
   repeat
      CurrentJob := Manager.TakeJob;
      while CurrentJob <> nil do
      begin
         RTLEventResetEvent (Manager.GotJob);
         CurrentJob();
         InterlockedDecrement (Manager.FJobs);
         RTLEventSetEvent (Manager.ExecutedJob);
         CurrentJob := Manager.TakeJob;
      end;
      { can leak events, so we provide a timeout }
      RtlEventWaitFor (Manager.GotJob, WORKER_IDLETIME);
   until Terminated;
end;

constructor TWorkManager.Create (NumWorkers: Integer);
var i: Integer;
begin
   GotJob := RTLEventCreate;
   ExecutedJob := RTLEventCreate;
   JobQueue := TJobQueue.Create;
   FJobs := 0;
   WorkerList := TWorkerList.Create;
   InitCriticalSection (ManagerLock);
   for i:=1 to NumWorkers do
      WorkerList.Add (TWorker.Create (self));
end;

procedure TWorkManager.AddJob (Job: TJob);
begin
   EnterCriticalSection (ManagerLock);
   try
      JobQueue.Add (Job);
      InterlockedIncrement (FJobs);
   finally LeaveCriticalSection (ManagerLock) end;
   RtlEventSetEvent (GotJob);
end;

function TWorkManager.TakeJob : TJob;
begin
   EnterCriticalSection (ManagerLock);
   try
      if JobQueue.Count > 0
      then begin
         Result := JobQueue.Last;
         JobQueue.Count := JobQueue.Count-1; // pop
      end
      else Result := nil;
   finally
      LeaveCriticalSection (ManagerLock);
   end;
end;

destructor TWorkManager.Destroy;
var Worker: TWorker;
begin
   if JobQueue.Count > 0 then
      raise (Exception.Create
                ('TWorkManager.Destroy: unfinished jobs'));
   for Worker in WorkerList do Worker.Terminate;
   for Worker in WorkerList do Worker.WaitFor;
   DoneCriticalSection (ManagerLock);
   RtlEventDestroy (GotJob);
   RtlEventDestroy (ExecutedJob);
   JobQueue.Destroy;
   WorkerList.Destroy;
end;

end.