unit workers_synch;
{ Job queue manager with fixed number of worker threads. Jobs have
  "report result" routines performed atomically from the main thread,
  afterwards jobs are freed. }

interface

uses {$ifdef unix}
  cthreads,
  {cmem,} // the c memory manager is on some systems much faster for
        // multi-threading
  {$endif}
  SysUtils, Classes, FGL;

const
   WORKER_IDLETIME = 100;
   MANAGER_IDLETIME = 500;
type
   TJob = class { Could be an interface }
      procedure Execute; virtual; abstract;
      procedure Report; virtual; abstract;
   end;
   TJobQueue = specialize TFPGList<TJob>;
   TWorkManager = class;
   
   TWorker = class (TThread)
   private
      Manager: TWorkManager;
   protected
      procedure Execute; override;
   public
      constructor Create (aManager: TWorkManager);
   end;
   TWorkerList = specialize TFPGList<TWorker>;
   
   TWorkManager = class
   private
      ManagerLock: TRTLCriticalSection;
      JobQueue: TJobQueue;
      ReportQueue: TJobQueue;
      WorkerList: TWorkerList;
      FJobs: Integer;
      GotJob: PRTLEvent;
      GotReport: PRTLEvent;
      function TakeJob (OldJob: TJob) : TJob;
   public
      constructor Create (NumWorkers: Integer);
      procedure AddJob (Job: TJob);
      procedure FinishJobs;
      destructor Destroy; override;
      property Jobs: Integer read FJobs;
      { procedure Suspend;
        procedure Resume; etc...}
   end;
   
implementation

constructor TWorker.Create (aManager: TWorkManager);
begin
   Manager := aManager;
   FreeOnTerminate := True;
   inherited Create (False); { start rightaway }
end;

procedure TWorker.Execute;
var CurrentJob : TJob;
begin
   repeat
      CurrentJob := Manager.TakeJob (nil);
      while CurrentJob <> nil do
      begin
         RTLEventResetEvent (Manager.GotJob);
         CurrentJob.Execute;
         CurrentJob := Manager.TakeJob (CurrentJob);
      end;
      { can leak events, so we provide a timeout }
      RtlEventWaitFor (Manager.GotJob, WORKER_IDLETIME);
   until Terminated;
end;

constructor TWorkManager.Create (NumWorkers: Integer);
var i: Integer;
begin
   GotJob := RTLEventCreate;
   GotReport := RTLEventCreate;
   JobQueue := TJobQueue.Create;
   ReportQueue := TJobQueue.Create;
   FJobs := 0;
   WorkerList := TWorkerList.Create;
   InitCriticalSection (ManagerLock);
   for i:=1 to NumWorkers do
      WorkerList.Add (TWorker.Create (self));
end;

procedure TWorkManager.AddJob (Job: TJob);
begin
   EnterCriticalSection (ManagerLock);
   try
      JobQueue.Add (Job);
      Inc (FJobs);
   finally
      LeaveCriticalSection (ManagerLock);
   end;
   RtlEventSetEvent (GotJob);
end;

function Pop (Queue: TJobQueue) : TJob;
begin
   if Queue.Count > 0
   then begin
      Result := Queue.Last;
      Queue.Count := Queue.Count-1; // pop
   end
   else Result := nil;
end;

function TWorkManager.TakeJob (OldJob: TJob) : TJob;
begin
   EnterCriticalSection (ManagerLock);
   try
      if OldJob <> nil then
      begin
         ReportQueue.Add (OldJob);
         Dec (FJobs);
         RtlEventSetEvent (GotReport);
      end;
      Result := Pop (JobQueue);
   finally
      LeaveCriticalSection (ManagerLock);
   end;
end;

procedure TWorkManager.FinishJobs;
var
   Job: TJob;
begin
   while (Jobs > 0) or (ReportQueue.Count > 0) do
   begin
      EnterCriticalSection (ManagerLock);
      try Job := Pop (ReportQueue);
      finally LeaveCriticalSection (ManagerLock) end;
      if Job <> nil
      then begin
         Job.Report;
         Job.Destroy;
      end else RtlEventWaitFor (GotReport, MANAGER_IDLETIME);
   end;  
end;

destructor TWorkManager.Destroy;
var Worker: TWorker;
begin
   if (Jobs > 0) or (ReportQueue.Count > 0) then
      raise (Exception.Create
                ('TWorkManager.Destroy: unfinished jobs or reports'));
   for Worker in WorkerList do Worker.Terminate;
   for Worker in WorkerList do Worker.WaitFor;
   DoneCriticalSection (ManagerLock);
   RtlEventDestroy (GotJob);
   RtlEventDestroy (GotReport);
   JobQueue.Destroy;
   ReportQueue.Destroy;
   WorkerList.Destroy;
end;

end.