unit workers_t_synch;
{ Job queue manager with fixed number of worker threads. Jobs have
  "report result" routines performed atomically from the main thread,
  afterwards jobs are freed. }

interface

uses {$ifdef unix}
  cthreads,
  {cmem,} // the c memory manager is on some systems much faster for
        // multi-threading
  {$endif}
  SysUtils, Classes, FGL;

const
   WORKER_IDLETIME = 100;
   MANAGER_IDLETIME = 500;
type
   TJob = class { Could be an interface }
      procedure Execute; virtual; abstract;
      procedure Report; virtual; abstract;
      function Info: String; virtual;
   end;
   TJobQueue = specialize TFPGList<TJob>;
   TWorkManager = class;
   
   TWorker = class (TThread)
   private
      ID: Integer;
      Manager: TWorkManager;
   protected
      procedure Execute; override;
   public
      constructor Create (aManager: TWorkManager; WorkerID: Integer);
      destructor Destroy; override;
   end;
   TWorkerList = specialize TFPGList<TWorker>;
   
   TWorkManager = class
   private
      ManagerLock: TRTLCriticalSection;
      JobQueue: TJobQueue;
      ReportQueue: TJobQueue;
      WorkerList: TWorkerList;
      FJobs: Integer;
      GotJob: PRTLEvent;
      GotReport: PRTLEvent;
      function TakeJob (OldJob: TJob) : TJob;
   public
      constructor Create (NumWorkers: Integer);
      procedure AddJob (Job: TJob);
      procedure FinishJobs;
      destructor Destroy; override;
      property Jobs: Integer read FJobs;
      { procedure Suspend;
        procedure Resume; etc...}
   end;
   
implementation

function TJob.Info: String;
begin
   Result := 'undefined';
end;

constructor TWorker.Create (aManager: TWorkManager; WorkerID: Integer);
begin
   WriteLn ('Creating worker ', WorkerID);
   ID := WorkerID;
   Manager := aManager;
   FreeOnTerminate := True;
   inherited Create (False); { start rightaway }
   WriteLn ('Worker ', WorkerID, ' created');
end;

procedure TWorker.Execute;
var CurrentJob : TJob;
begin
   WriteLn ('Executing worker ', ID);
   repeat
      CurrentJob := Manager.TakeJob (nil);
      while CurrentJob <> nil do
      begin
         WriteLn ('Worker ', ID, ': processing ', CurrentJob.Info);
         RTLEventResetEvent (Manager.GotJob);
         CurrentJob.Execute;
         WriteLn ('Worker ', ID, ': processed ', CurrentJob.Info);
         CurrentJob := Manager.TakeJob (CurrentJob);
      end;
      { can leak events, so we provide a timeout }
      RtlEventWaitFor (Manager.GotJob, WORKER_IDLETIME);
   until Terminated;
end;

destructor TWorker.Destroy;
begin
   WriteLn ('Worker ', ID, ' Destroy');
   inherited Destroy;
end;

constructor TWorkManager.Create (NumWorkers: Integer);
var i: Integer;
begin
   WriteLn ('Creating WorkManager');
   GotJob := RTLEventCreate;
   GotReport := RTLEventCreate;
   JobQueue := TJobQueue.Create;
   ReportQueue := TJobQueue.Create;
   FJobs := 0;
   WorkerList := TWorkerList.Create;
   InitCriticalSection (ManagerLock);
   for i:=1 to NumWorkers do
      WorkerList.Add (TWorker.Create (self, i));
end;

procedure TWorkManager.AddJob (Job: TJob);
begin
   EnterCriticalSection (ManagerLock);
   try
      WriteLn ('WorkManager: adding ', Job.Info);
      JobQueue.Add (Job);
      Inc (FJobs);
   finally
      LeaveCriticalSection (ManagerLock);
   end;
   RtlEventSetEvent (GotJob);
end;

function Pop (Queue: TJobQueue; Info: String) : TJob;
begin
   if Queue.Count > 0
   then begin
      Result := Queue.Last;
      WriteLn (Info,': popping ', Queue.Last.Info,
               ' of a queue of ', Queue.Count, ' jobs');
      Queue.Count := Queue.Count-1; // pop
   end
   else begin
      WriteLn (Info, ': no job to pop');
      Result := nil;
   end
end;

function TWorkManager.TakeJob (OldJob: TJob) : TJob;
begin
   EnterCriticalSection (ManagerLock);
   try
      if OldJob <> nil then
      begin
         ReportQueue.Add (OldJob);
         Dec (FJobs);
         WriteLn ('WorkManager: queue ', OldJob.Info,
                  ' for report, remaining ', FJobs, ' jobs');
         RtlEventSetEvent (GotReport);
      end;
      Result := Pop (JobQueue, 'TakeJob');
   finally
      LeaveCriticalSection (ManagerLock);
   end;
end;

procedure TWorkManager.FinishJobs;
var
   Job: TJob;
begin
   WriteLn ('WorkManager: FinishJobs');
   while (Jobs > 0) or (ReportQueue.Count > 0) do
   begin
      WriteLn ('WorkManager: Waiting for finish');
      EnterCriticalSection (ManagerLock);
      try Job := Pop (ReportQueue, 'Finish report');
      finally LeaveCriticalSection (ManagerLock) end;
      if Job <> nil
      then begin
         Job.Report;
         Job.Destroy;
      end else RtlEventWaitFor (GotReport, MANAGER_IDLETIME);
   end;  
   WriteLn ('WorkManager: finished');
end;

destructor TWorkManager.Destroy;
var Worker: TWorker;
begin
   if (Jobs > 0) or (ReportQueue.Count > 0) then
      raise (Exception.Create
                ('TWorkManager.Destroy: unfinished jobs or reports'));
   WriteLn ('WorkManager.Destroy: sending termination signals');
   for Worker in WorkerList do Worker.Terminate;
   WriteLn ('WorkManager.Destroy: waiting for termination');
   for Worker in WorkerList do Worker.WaitFor;
   WriteLn ('WorkManager.Destroy: free resoruces');
   DoneCriticalSection (ManagerLock);
   RtlEventDestroy (GotJob);
   RtlEventDestroy (GotReport);
   JobQueue.Destroy;
   ReportQueue.Destroy;
   WorkerList.Destroy;
   WriteLn ('WorkManager.Destroy: end');
end;

end.